(* File: utils.mli

    Copyright (C) 2007-

      Jane Street Holding, LLC
      Author: Markus Mottl
      email: mmottl\@janestcapital.com
      WWW: http://www.janestcapital.com/ocaml

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(** Utils: utility functions for user convenience *)

open Bigarray
open Common
open Read_ml
open Type_class

(* CRv2 sweeks: I'm dubious that writing out the header length using a
   variable-sized encoding is ever necessary.  Perhaps the `Variable possibility
   should be removed.  And do we ever need to write out without a length?  If
   not, perhaps we should remove the ?header argument and always write fixed
   size length header.

   mmottl: It's true that for most applications there is not much benefit
   to using variable-sized headers.  They safe a little bit of space,
   which may be non-negligible when lots of very small messages are
   stored in a file.  They are also slightly more efficient to read
   unless the data is embedded in a stream.  I'd say I'm in favor of
   fixed-size headers, but I'd like to keep the option for variable-sized
   ones around.  Concerning writing without size-information: if you
   only store one value in a file, the file size determines the length.
   Furthermore, the size is really only needed for efficiency reasons
   (you know how much to buffer before even attempting unmarshalling),
   because the binary protocol is self-delimiting.

   sweeks: The space savings seems possibly important.  Although, if one were
   writing out messages of one type to a file and really wanted to save space,
   wouldn't you use no header (because self-delimiting)?  Perhaps we should
   have fixed-size header or no header, but not variable sized?  Also, it seems
   like the default should be fixed-size header, since that is the common case.
   Or perhaps, to fight against mismatch errors, make the header argument
   required.

   Let's leave this as a CRv2 in 102 and talk about it later.
*)
val bin_dump : ?header : [ `Fixed64 | `Variable ] -> 'a writer -> 'a -> buf
(** [bin_dump ?header writer v] uses [writer] to first compute the size of
    [v] in the binary protocol, then allocates a buffer of exactly this
    size, and then writes out the value.  If [header] is defined, the
    size of the resulting binary string will be prefixed.  If [header]
    is [`Fixed64], this size information will be exactly 8 bytes long
    (exact bit representation of the 64bit integer encoding the size),
    otherwise, if it is [`Variable], a variable size encoding ([Nat0.t])
    will be used.

    @return the buffer containing the written out value.

    @param header = nothing

    @raise Failure if the size of the value changes during writing,
    and any other exceptions that the binary writer in [writer] can raise.
*)

(* sweeks: I commented out [bin_read_stream] because we don't need it in 102. *)
val bin_read_stream :
  ?fixed_header : bool ->
  ?max_size : int ->
  read : (buf -> pos : int -> len : int -> unit) ->
  'a reader -> 'a
(** [bin_read_stream ?fixed_header ?max_size ~read reader] reads binary
    protocol data from a stream as generated by the [read] function,
    which places data of a given length into a given buffer.  Uses a
    variably-sized header if [fixed_header] is [false], a fixed-size
    one otherwise.  The [reader] type class will be used for conversion
    to OCaml-values.

    @param fixed_header = false
    @param max_size = nothing

    @raise Failure if the size of the value disagrees with the one
    specified in the header, and any other exceptions that the binary
    reader associated with [reader] can raise.

    @raise Failure if the size reported in the data header is longer than
    [max_size].
*)


(** Read buffers for efficient parsing of [Nat0.t] size-prefixed values *)
module Read_buf : sig
  type t (** Type of read buffer *)

  val create : unit -> t
  (** [create ()] @return a read buffer. *)

  val next : ?max_size : int -> 'a reader -> (t -> 'a option) Read_ml.reader
  (** [next ?max_size reader buf ~pos_ref rbuf] looks at read buffer
      [rbuf] and (ordinary) buffer [buf] starting at the position stored
      in [pos_ref] to parse OCaml-values encoded in the size-prefixed
      binary protocol (variable size prefix only!).

      NOTE: [buf] must not contain trailing garbage after messages
      (including partial ones)!  Use {!Bigarray.Array1.sub} to constrain
      the buffer to the exact range covered by valid data.

      The function first attempts to read the size (binary protocol
      natural number) of a value to be parsed, then the value using
      [reader].  If [buf] does not contain enough data to read the size
      and/or the value, a partial read will be performed.  [rbuf] will
      then buffer all available data, thus always making it safe to store
      other data in [buf] as soon as {!next} returns.  Subsequent calls
      to {!next} with additional data allows one to continue parsing.

      [max_size] determines the maximum binary-encoded size of a value
      allowed.  [pos_ref] will contain the position after the last
      character read in [buf].

      @return [Some value] if a value was read, [None] otherwise,
      or raise any exceptions associated with parse errors in [reader].

      @raise Failure if [max_size] is exceeded or if the protocol lies
      about the binary-encoded size of the value.

      @param max_size default = None
  *)

  val enforce_buf_size : t -> int -> bool
  (** [enforce_buf_size rbuf n] enforces a maximum size of the internal
      buffer to be [n] bytes IF POSSIBLE.  If the currently buffered
      data (and hence the buffer) is larger than that, or if the buffer
      is smaller, or if [n] is smaller than the minimum internal buffer
      size, nothing will happen.  Use {!purge} to forget buffered data.
      @return [true] if the size of the internal buffer now fits into [n],
      [false] otherwise.
  *)

  val purge : t -> unit
  (** [purge rbuf] empties [rbuf], thus forgetting about previously
      buffered data.  The internal buffer may still remain large.
      Use {!enforce_buf_size} to change that. *)

  val get_size : t -> int
  (** [get_size rbuf] @return the size of the internal buffer of [rbuf]. *)

  val get_buffered_data_size : t -> int
  (** [get_buffered_data_size rbuf] @return the size of buffered data in
      [rbuf], including partially read length prefixes. *)

  val get_expected_data_size : t -> int option
  (** [get_expected_data_size rbuf] @return [Some expected_size] if the
      size header has already been read, where [expected_size] is the
      total size of the binary-encoded value without, or return [None]
      if this information is not yet available. *)
end


(** Conversion of binable types *)

module type Make_binable_spec = sig
  type t
  type binable

  val to_binable : t -> binable
  val of_binable : binable -> t

  val bin_size_binable : binable Size.sizer
  val bin_write_binable : binable Map_to_safe.writer
  val bin_write_binable_ : binable Unsafe_write_c.writer
  val bin_read_binable : binable Read_ml.reader
  val bin_read_binable_ : binable Unsafe_read_c.reader
  val bin_read_binable__ : (int -> binable) Unsafe_read_c.reader
  val bin_writer_binable : binable Type_class.writer
  val bin_reader_binable : binable Type_class.reader
  val bin_binable : binable Type_class.t
end

module Make_binable (Bin_spec : Make_binable_spec) : sig
  open Bin_spec

  type t = Bin_spec.t

  val bin_size_t : t Size.sizer
  val bin_write_t : t Map_to_safe.writer
  val bin_write_t_ : t Unsafe_write_c.writer
  val bin_read_t : t Read_ml.reader
  val bin_read_t_ : t Unsafe_read_c.reader
  val bin_read_t__ : (int -> t) Unsafe_read_c.reader
  val bin_writer_t : t Type_class.writer
  val bin_reader_t : t Type_class.reader
  val bin_t : t Type_class.t
end
